#!/usr/bin/env python
""" Compile CUDA source code and setup Python package 'nipet'
    for namespace package 'niftypet'.
"""
__author__      = "Pawel J. Markiewicz"
__copyright__   = "Copyright 2018"
# ---------------------------------------------------------------------------------

from setuptools import setup, find_packages

import os
import sys
import platform
from subprocess import call, Popen, PIPE

if 'DISPLAY' in os.environ:
    from Tkinter import Tk
    from tkFileDialog import askdirectory

import cudasetup_hdr as cs

#-------------------------------------------------------------------------
# The below function is a copy of the same function in install_tools.py
# in NIMPA
def update_resources(Cnt):
    '''Update resources.py with the paths to the new installed apps.
    '''
    # list of path names which will be saved
    key_list = ['PATHTOOLS', 'RESPATH', 'REGPATH', 'DCM2NIIX', 'HMUDIR']

    # get the local path to NiftyPET resources.py
    path_resources = cs.path_niftypet_local()
    resources_file = os.path.join(path_resources,'resources.py')

    # update resources.py
    if os.path.isfile(resources_file):
        f = open(resources_file, 'r')
        rsrc = f.read()
        f.close()
        # get the region of keeping in synch with Python
        i0 = rsrc.find('### start NiftyPET tools ###')
        i1 = rsrc.find('### end NiftyPET tools ###')
        pth_list = []
        for k in key_list:
            if k in Cnt:
                pth_list.append('\'' + Cnt[k].replace("\\","/") + '\'')
            else:
                pth_list.append('\'\'')

        # modify resources.py with the new paths
        strNew = '### start NiftyPET tools ###\n'
        for i in range(len(key_list)):
            if pth_list[i] != '\'\'':
                strNew += key_list[i]+' = '+pth_list[i] + '\n'
        rsrcNew = rsrc[:i0] + strNew + rsrc[i1:]
        f = open(resources_file, 'w')
        f.write(rsrcNew)
        f.close()

    return Cnt
#-------------------------------------------------------------------------

if not 'Windows' in platform.system() and not 'Linux' in platform.system():
    print 'e> the current operating system is not supported.'
    raise SystemError('OS: Unknown Sysytem.')

#----------------------------------------------------
# select the supported GPU device and install resources.py
print ' '
print '---------------------------------------------'
print 'i> setting up CUDA ...'
gpuarch = cs.resources_setup()
#----------------------------------------------------



#===============================================================
# Hardware mu-maps
print ' '
print '---------------------------------------------'
print 'i> indicate the location of hardware mu-maps:'

#---------------------------------------------------------------
# get the local path to NiftyPET resources.py
path_resources = cs.path_niftypet_local()
# if exists, import the resources and get the constants
if os.path.isfile(os.path.join(path_resources,'resources.py')):
    sys.path.append(path_resources)
    try:
        import resources
    except ImportError as ie:
        print '---------------------------------------------------------------------------------'
        print 'e> Import Error: NiftyPET''s resources file <resources.py> could not be imported.'
        print '---------------------------------------------------------------------------------'
        raise SystemError('Missing resources file')
    # get the current setup, if any
    Cnt = resources.get_setup()

    # assume the hardware mu-maps are not installed
    hmu_flg = False
    # go through each piece of the hardware components
    if 'HMUDIR' in Cnt and Cnt['HMUDIR']!='':
        for hi in Cnt['HMULIST']:
            if os.path.isfile(os.path.join(Cnt['HMUDIR'],hi)):
                hmu_flg = True
            else:
                hmu_flg = False
                break
    # if not installed ask for the folder through GUI
    # otherwise the path will have to be filled manually
    if not hmu_flg and 'DISPLAY' in os.environ:
        Tk().withdraw()
        Cnt['HMUDIR'] = askdirectory(
            title='Folder for hardware mu-maps',
            initialdir=os.path.expanduser('~')
        )
    #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # update the path in resources.py
    update_resources(Cnt)
    #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
else:
    raise SystemError('Missing file: resources.py')

print '--------------------------------------'
print 'i> hardware mu-maps have been located.'
print '--------------------------------------'
#===============================================================



#===============================================================
print '---------------------------------'
print 'i> CUDA compilation for NIPET ...'
print '---------------------------------'

path_current = os.path.dirname( os.path.realpath(__file__) )
path_build = os.path.join(path_current, 'build')
if not os.path.isdir(path_build): os.makedirs(path_build)
os.chdir(path_build)

# cmake installation commands
cmd = []
cmd.append([
    'cmake',
    os.path.join('..','niftypet'),
    '-DPYTHON_INCLUDE_DIRS='+cs.pyhdr,
    '-DPYTHON_PREFIX_PATH='+cs.prefix,
    '-DCUDA_NVCC_FLAGS='+gpuarch
])
cmd.append(['cmake', '--build', './'])

if platform.system()=='Windows':
    cmd[0] += ['-G', Cnt['MSVC_VRSN']]
    cmd[1] += ['--config', 'Release']

# error string for later reporting
errstr = []
# the log files the cmake results are written
cmakelog = ['py_cmake_config.log', 'py_cmake_build.log']
# run commands with logging
for ci in range(len(cmd)):
    with open(cmakelog[ci], 'w') as f:
        p = Popen(cmd[ci], stdout=PIPE, stderr=PIPE)
        for c in iter(lambda: p.stdout.read(1), ''):
            sys.stdout.write(c)
            f.write(c)
    # get the pipes outputs
    stdout, stderr = p.communicate()
    ei = stderr.find('error')
    if ei>=0:
        errstr.append(stderr[ei:ei+60]+'...')
    else:
        errstr.append('_')

    if stderr:
        print 'c>-------- reports -----------'
        print stderr+'c>------------ end ---------------'

    print ' '
    print stdout


print ' '
print '--- error report ---'
for ci in range(len(cmd)):
    if errstr[ci] != '_':
        print 'e> found error(s) in ', ' '.join(cmd[ci]), '>>', errstr[ci]
        print ' '
print '--- end ---'

# come back from build folder
os.chdir(path_current)
#===============================================================





#===============================================================
# PYTHON SETUP
#===============================================================

print 'i> found those packages:'
print find_packages(exclude=['docs'])

with open('README.rst') as file:
    long_description = file.read()

#---- for setup logging -----
stdout = sys.stdout
stderr = sys.stderr
log_file = open('setup_nipet.log', 'w')
sys.stdout = log_file
sys.stderr = log_file
#----------------------------

if platform.system() == 'Linux' :
    fex = '*.so'
elif platform.system() == 'Windows' :
    fex = '*.pyd'
#----------------------------
setup(
    name='nipet',
    license = 'Apache 2.0',
    version='1.1.14',
    description='CUDA-accelerated Python utilities for high-throughput PET/MR image reconstruction and analysis.',
    long_description=long_description,
    author='Pawel J. Markiewicz',
    author_email='p.markiewicz@ucl.ac.uk',
    url='https://github.com/pjmark/NiftyPET',
    keywords='PET image reconstruction and analysis',
    install_requires=['nimpa>=1.1.0', 'pydicom>=1.0.2,<=1.2.2',
      'nibabel>=2.2.1, <=2.3.1', 'tqdm>=4.27'],
    packages=find_packages(exclude=['docs']),
    package_data={
        'niftypet': ['auxdata/*'],
        'niftypet.nipet.dinf': [fex],
        'niftypet.nipet.lm' : [fex],
        'niftypet.nipet.prj' : [fex],
        'niftypet.nipet.sct' : [fex],
        'niftypet.nipet' : [fex],
    },
    zip_safe=False,
    # classifiers=[
    #     'Development Status :: 5 - Production/Stable',
    #     'Intended Audience :: Science/Research',
    #     'Intended Audience :: Healthcare Industry'
    #     'Programming Language :: Python :: 2.7',
    #     'License :: OSI Approved :: Apache Software License',
    #     'Operating System :: POSIX :: Linux',
    #     'Programming Language :: C',
    #     'Topic :: Scientific/Engineering :: Medical Science Apps.'
    # ],
    # namespace_packages=['niftypet'],
)
#===============================================================



"""hist.py: processing of PET list-mode data: histogramming and randoms estimation."""
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"
#-------------------------------------------------------------------------------
import numpy as np
from math import pi
import sys
import os
import scipy.ndimage as ndi
import nibabel as nib
import cPickle as pickle

#CUDA extension module
import mmr_lmproc

#================================================================================
# HISTOGRAM THE LIST-MODE DATA
#--------------------------------------------------------------------------------
def mmrhist(
        datain,
        scanner_params,
        t0=0, t1=0,
        outpath='',
        frms=np.array([0], dtype=np.uint16),
        use_stored=False,
        store=False,
        cmass_sig=5):
    '''Process the list-mode data and return histogram, head curves, and centre of mass for motion detection.
    '''

    # constants, transaxial and axial LUTs are extracted
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    hst = hist(
            datain, txLUT, axLUT, Cnt,
            frms=frms,
            use_stored=use_stored,
            store=store,
            outpath=outpath,
            t0=t0, t1=t1,
            cmass_sig=cmass_sig)

    return hst


def hist(datain, txLUT, axLUT, Cnt, frms=np.array([0], dtype=np.uint16), use_stored=False, store=False, outpath='', t0=0, t1=0, cmass_sig=5 ):

    # histogramming with bootstrapping:
    # Cnt['BTP'] = 0: no bootstrapping [default];
    # Cnt['BTP'] = 1: non-parametric bootstrapping;
    # Cnt['BTP'] = 2: parametric bootstrapping (using Poisson distribution with mean = 1)

    #number of dynamic frames
    nfrm = len(frms)
    if    Cnt['SPN']==1:  nsinos=Cnt['NSN1']
    elif  Cnt['SPN']==11: nsinos=Cnt['NSN11']
    elif  Cnt['SPN']==0:  nsinos=Cnt['NSEG0']

    if Cnt['VERBOSE']: print 'i> histograming with span', Cnt['SPN'], 'and', nfrm, 'dynamic frames.'


    if use_stored==True and 'sinos' in datain and os.path.basename(datain['sinos'])=='sinos_s'+str(Cnt['SPN'])+'_n'+str(nfrm)+'_frm-'+str(t0)+'-'+str(t1)+'.npy' :

        # nele, ttags, tpos = mmr_lmproc.lminfo(datain['lm_bf'])
        # nitag = (ttags[1]-ttags[0]+999)/1000

        hstout = {}
        (hstout['phc'], hstout['dhc'], hstout['mss'], hstout['pvs'],
         hstout['bck'], hstout['fan'], hstout['psn'], hstout['dsn'],
         hstout['ssr']) =   np.load( datain['sinos'] )

        nitag = len(hstout['phc'])
        if Cnt['VERBOSE']: print 'i> duration by integrating time tags [s]:', nitag

    elif os.path.isfile(datain['lm_bf']):
        # gather info about the LM time tags
        nele, ttags, tpos = mmr_lmproc.lminfo(datain['lm_bf'])
        nitag = (ttags[1]-ttags[0]+999)/1000
        if Cnt['VERBOSE']: print 'i> duration by integrating time tags [s]:', nitag

        # adjust frame time if outside the limit
        if t1>nitag: t1 = nitag
        # check if the time point is allowed
        if t0>=nitag:
            print 'e> time frame definition outside the list-mode data acquisition time!'
            raise ValueError('Not allowed time frame definition')

        # ---------------------------------------
        # preallocate all the output arrays
        VTIME = 2
        MXNITAG = 5400 #limit to 1hr and 30mins
        if (nitag>MXNITAG): 
            tn = MXNITAG/(1<<VTIME)
        else: 
            tn = (nitag+(1<<VTIME)-1)/(1<<VTIME)

        pvs = np.zeros((tn, Cnt['NSEG0'], Cnt['NSBINS']), dtype=np.uint32)
        phc = np.zeros((nitag), dtype=np.uint32)
        dhc = np.zeros((nitag), dtype=np.uint32)
        mss = np.zeros((nitag), dtype=np.float32)

        bck = np.zeros((2, nitag, Cnt['NBCKT']), dtype=np.uint32)
        fan = np.zeros((nfrm, Cnt['NRNG'], Cnt['NCRS']), dtype=np.uint32)

        if nfrm==1:
            psino = np.zeros((nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.uint16)
            dsino = np.zeros((nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.uint16)
        elif nfrm>1:
            psino = np.zeros((nfrm, nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.uint8)
            dsino = np.zeros((nfrm, nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.uint8)

        # single slice rebinned prompots
        ssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.uint32)

        hstout = {
            'phc':phc,
            'dhc':dhc,
            'mss':mss,
            'pvs':pvs,

            'bck':bck,
            'fan':fan,

            'psn':psino,
            'dsn':dsino,
            'ssr':ssr

        }
        # ---------------------------------------

        # do the histogramming and processing
        mmr_lmproc.hist(
                    hstout,
                    datain['lm_bf'],
                    frms, 
                    t0, t1,
                    txLUT, axLUT, Cnt)
        if store:
            if outpath=='':
                fsino = os.path.dirname(datain['lm_bf'])
            else:
                fsino = os.path.join(outpath, 'sino')
                nipet.mmraux.create_dir(fsino)
            # complete the path with the file name
            fsino = os.path.join(fsino, 'sinos_s'+str(Cnt['SPN'])+'_n'+str(nfrm)+'_frm-'+str(t0)+'-'+str(t1)+'.npy')
            # store to the above path
            np.save(fsino,  (hstout['phc'], hstout['dhc'], hstout['mss'], hstout['pvs'],
                    hstout['bck'], hstout['fan'], hstout['psn'], hstout['dsn'], hstout['ssr']))

    else:
        print 'e> input list-mode data not defined.'
        return

    #short (interval) projection views
    pvs_sgtl = np.right_shift(hstout['pvs'], 8).astype(np.float32)
    pvs_crnl = np.bitwise_and(hstout['pvs'], 255).astype(np.float32)

    cmass = Cnt['SO_VXZ']*ndi.filters.gaussian_filter(hstout['mss'], cmass_sig, mode='mirror')
    if Cnt['VERBOSE']: print 'i> centre of mass of axial radiodistribution (filtered with Gaussian of SD =', cmass_sig, '):  COMPLETED.'

    #========================== BUCKET SINGLES ==============================
    #number of single rates reported for the given second
    nsr = (hstout['bck'][1,:,:]>>30)
    #average in a second period
    hstout['bck'][0,nsr>0] /= nsr[nsr>0]
    #time indeces when single rates given
    tmsk = np.sum(nsr,axis=1)>0
    single_rate = np.copy(hstout['bck'][0,tmsk,:])
    #time
    t = np.arange(nitag)
    t = t[tmsk]
    #get the average bucket singles:
    buckets = np.int32( np.sum(single_rate,axis=0)/single_rate.shape[0] )
    if Cnt['VERBOSE']: print 'i> dynamic and static buckets single rates:  COMPLETED.'
    #=========================================================================

    # account for the fact that when t0==t1 that means that full dataset is processed
    if t0==t1: t1 = t0+nitag

    pdata={
        't0':t0,
        't1':t1,
        'dur':t1-t0,                #duration
        'phc':hstout['phc'],        #prompts head curve
        'dhc':hstout['dhc'],        #delayeds head curve
        'cmass':cmass,              #centre of mass of the radiodistribution in axial direction
        'pvs_sgtl':pvs_sgtl,        #sagittal projection views in short intervals
        'pvs_crnl':pvs_crnl,        #coronal projection views in short intervals
        
        'fansums':hstout['fan'],    #fan sums of delayeds for variance reduction of random event sinograms
        'sngl_rate':single_rate,    #bucket singles over time
        'tsngl':t,                  #time points of singles measurements in list-mode data
        'buckets':buckets,          #average bucket singles

        'psino':hstout['psn'],      #prompt sinogram
        'dsino':hstout['dsn'],      #delayeds sinogram
        'pssr' :hstout['ssr']       #single-slice rebinned sinogram of prompts
    }

    return pdata

#================================================================================
# GET REDUCED VARIANCE RANDOMS
#--------------------------------------------------------------------------------
def rand(fansums, txLUT, axLUT, Cnt):

    if    Cnt['SPN']==1:  nsinos=Cnt['NSN1']
    elif  Cnt['SPN']==11: nsinos=Cnt['NSN11']
    elif  Cnt['SPN']==0:  nsinos=Cnt['NSEG0']

    #number of frames
    nfrm = fansums.shape[0]
    if Cnt['VERBOSE']: print 'i> # of dynamic frames:', nfrm

    #random sino and estimated crystal map of singles put into a dictionary
    rsn  = np.zeros((nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
    cmap = np.zeros((Cnt['NCRS'], Cnt['NRNG']), dtype=np.float32)
    rndout = {
        'rsn': rsn,
        'cmap':cmap,
    }

    #save results for each frame
    
    rsino = np.zeros((nfrm, nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
    crmap = np.zeros((nfrm, Cnt['NCRS'], Cnt['NRNG']), dtype=np.float32)
    
    for i in range(nfrm):
        rndout['rsn'][:,:,:] = 0
        rndout['cmap'][:,:]  = 0
        mmr_lmproc.rand(rndout, fansums[i,:,:], txLUT, axLUT, Cnt)
        rsino[i,:,:,:] = rndout['rsn']
        crmap[i,:,:] = rndout['cmap']

    if nfrm==1:
        rsino = rsino[0,:,:,:]
        crmap = crmap[0,:,:]

    return rsino, crmap


#================================================================================
# NEW!! GET REDUCED VARIANCE RANDOMS (BASED ON PROMPTS)
#--------------------------------------------------------------------------------
def prand(fansums, pmsk, txLUT, axLUT, Cnt):

    if    Cnt['SPN']==1:  nsinos=Cnt['NSN1']
    elif  Cnt['SPN']==11: nsinos=Cnt['NSN11']
    elif  Cnt['SPN']==0:  nsinos=Cnt['NSEG0']

    #number of frames
    nfrm = fansums.shape[0]
    if Cnt['VERBOSE']: print 'i> # of dynamic frames:', nfrm

    #random sino and estimated crystal map of singles put into a dictionary
    rsn  = np.zeros((nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
    cmap = np.zeros((Cnt['NCRS'], Cnt['NRNG']), dtype=np.float32)
    rndout = {
        'rsn': rsn,
        'cmap':cmap,
    }

    #save results for each frame
    
    rsino = np.zeros((nfrm, nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
    crmap = np.zeros((nfrm, Cnt['NCRS'], Cnt['NRNG']), dtype=np.float32)
        
    for i in range(nfrm):
        rndout['rsn'][:,:,:] = 0
        rndout['cmap'][:,:]  = 0
        mmr_lmproc.prand(rndout, pmsk, fansums[i,:,:], txLUT, axLUT, Cnt)
        rsino[i,:,:,:] = rndout['rsn']
        crmap[i,:,:] = rndout['cmap']

    if nfrm==1:
        rsino = rsino[0,:,:,:]
        crmap = crmap[0,:,:]

    return rsino, crmap







#================================================================================
def sino2nii(sino, Cnt, fpth):
    '''save sinogram in span-11 into NIfTI file'''
    #number of segments
    segn = len(Cnt['SEG'])
    cumseg = np.cumsum(Cnt['SEG'])
    cumseg = np.append([0], cumseg)

    #plane offset (relative to 127 planes of seg 0) for each segment
    OFF = np.min( abs( np.append([Cnt['MNRD']], [Cnt['MXRD']], axis=0) ), axis=0 )
    niisn = np.zeros(( Cnt['SEG'][0], Cnt['NSANGLES'], Cnt['NSBINS'], segn), dtype=sino.dtype)

    #first segment (with direct planes)
    # tmp = 
    niisn[:,:,:,0] = sino[Cnt['SEG'][0]-1::-1, ::-1, ::-1]

    for iseg in range(1,segn):
        niisn[OFF[iseg]:OFF[iseg]+Cnt['SEG'][iseg], :, :, iseg] = sino[cumseg[iseg]+Cnt['SEG'][iseg]-1:cumseg[iseg]-1:-1, ::-1, ::-1 ]

    niisn = np.transpose(niisn, (2, 1, 0, 3))

    nim = nib.Nifti1Image(niisn, np.eye(4))
    nib.save(nim, fpth)


#=================================================================================
# create michelogram map for emission data, only when the input sino in in span-1
def get_michem(sino, axLUT, Cnt):
    # span:
    spn = -1

    if Cnt['SPN']==1:
        slut = np.arange(Cnt['NSN1']) #for span 1, one-to-one mapping
    elif Cnt['SPN']==11:
        slut = axLUT['sn1_sn11']
    else:
        print 'e> sino is not in span-1 neither span-11'
        sys.exit()

    #acitivity michelogram
    Mem = np.zeros((Cnt['NRNG'],Cnt['NRNG']), dtype=np.float32)
    #sino to ring number & sino-1 to sino-11 index:
    sn1_rno  = axLUT['sn1_rno']
    #sum all the sinograms inside 
    ssm = np.sum(sino, axis=(1,2))

    for sni in range(len(sn1_rno)):
        r0 = sn1_rno[sni,0]
        r1 = sn1_rno[sni,1]
        Mem[r1,r0] = ssm[slut[sni]]

    return Mem


#=================================================================================
#---------------------------------------------------------------------------------
#=================================================================================


#------------------------------------------------------
def draw_frames(hst, tfrms, plot_diff=True):
    import matplotlib.pyplot as plt
    diff = np.int64(hst['phc']) - np.int64(hst['dhc'])
    plt.figure()
    plt.plot(hst['phc'], label='prompts')
    plt.plot(hst['dhc'], label='randoms')
    if plot_diff:
        plt.plot(diff, label='difference')

    K = [f[0] for f in tfrms if isinstance(f,list)]
    for k in K:
        yval = hst['phc'][k]
        if yval<0.2*np.max(hst['phc']):
            yval = 0.2*np.max(hst['phc'])
        plt.plot([k, k], [0, yval], 'k--', lw=.75)
    plt.legend()
    plt.xlabel('time [sec]')
    plt.ylabel('counts/sec')
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
#------------------------------------------------------


def get_time_offset(hst):
    '''
    Detects when the signal is stronger than the randoms (noise) in the list-mode data stream.
    '''
    # detect when the signal (here prompt data) is almost as strong as randoms
    s = hst['dhc']>0.98*hst['phc']
    # return index, which will constitute time in seconds, for this offset
    return  len(s)-np.argmax(s[::-1])-1

def split_frames(hst, Tref=0, t0=0, t1=0):
    '''
    Splits the whole acquisition data into approximately statistically 
    equivalent frames relative to the reference frame whose duration is 
    Tref or t1-t0.  The next frames will have a similar count level.
    hst: histogram dictionary
    Tref: reference duration in seconds
    t0: start time of the reference frame
    t1: end time of the reference frame
    '''
    # get the offset
    toff = get_time_offset(hst)
    # difference between prompts and randoms
    diff = np.int64(hst['phc']) - np.int64(hst['dhc'])

    # follow up index
    i = t0 + (toff)*(t0<=0)
    if Tref>0:
        j = i+Tref
    elif t1>0:
        j = t1 + (toff)*(t0<=0)
    else:
        raise ValueError('e> could not figure out the reference frame.')

    # reference count level
    cref = np.sum(diff[i:j])
    # cumulative sum of the difference
    csum = np.cumsum(diff)

    i = 0
    j = toff
    # threshold to be achieved
    thrsh = csum[j-1] + cref
    fdur = [toff]
    frms = ['timings', [0, toff]]
    clvl = [0]
    print 'counts t(%d,%d) = %d. diff=%d' % ( i,j,clvl[-1] , np.sum(diff[i:j])-cref )
    while thrsh<csum[-1]:
        i = j
        j = np.argmax(csum>thrsh)
        fdur.append(j-i)
        frms.append([i,j])
        clvl.append(np.sum(diff[i:j]))
        print 'counts t(%d,%d) = %d. diff=%d' % ( i,j,clvl[-1] , np.sum(diff[i:j])-cref )
        thrsh += cref
    # last remianing frame
    i=j
    j=hst['dur']
    # if last frame is short, include it in the last one.
    if np.sum(diff[i:])>.5*cref:
        fdur.append(j-i)
        frms.append([i,j])
        clvl.append(np.sum(diff[i:]))
    else:
        fdur[-1] += j-i
        frms[-1][-1] += j-i
        clvl[-1] += np.sum(diff[i:])
        i = frms[-1][0]
    print 'counts t(%d,%d) = %d. diff=%d' % ( i,j,clvl[-1] , np.sum(diff[i:j])-cref )
    return {'timings':frms, 'fdur':fdur, 'fcnts':clvl, 'offset':toff, 'csum':csum}


#-------------------------------------------------------------------------------------------------

def frame_position(hst, tposition, Cref=0, tr0=0, tr1=15, verbose = True):
    ''' hst: histogram data
        tposition: time position (middle point) of the frame to be defined
        Cref: reference count level to be in the frame (prompts - delays)
        tr0, tr1: time definitions of reference frame whose count level
        will be used as the reference Cref.  If Cref is not defined (i.e. = 0)
        then the tr0 and tr1 will be used.
    '''
    
    # claculate the difference between the prompts and delays (more realistic count level)
    diff = np.int64(hst['phc']) - np.int64(hst['dhc'])
    # cumulative sum for calculating count levels in arbitrary time windows
    cumdiff = np.cumsum(diff)

    if Cref==0:
        Cref = cumdiff[tr1]-cumdiff[tr0-1]

    if Cref<0:
        raise ValueError('The reference count level has to be non-negative')

    if verbose: print 'i> reference count level:', Cref


    stp0 = 0
    stp1 = 0
    Cw = 0
    while Cw<Cref:
        # check if it is possible to widen the sampling window both ways
        if (tposition-stp0-1)>0: stp0 += 1
        if (tposition+stp1+1)<=len(cumdiff)-1: stp1 += 1
        Cw = cumdiff[tposition+stp1] - cumdiff[tposition-stp0-1]

    tw0 = tposition-stp0
    tw1 = tposition+stp1
    Tw = tw1 - tw0
    if verbose:
        print 'i> time window t[{}, {}] of duration T={} and count level Cw={}'.format(tw0, tw1, Tw, Cw)

    return (tw0, tw1)


def auxilary_frames(hst, t_frms, Cref=0, tr0=0, tr1=15, verbose = True):
    ''' Get auxiliary time frames with equal count levels for constant precision in
        the estimation of subject motion based on PET data. 
    '''
    
    # calculate the difference between the prompts and delays (more realistic count level)
    diff = np.int64(hst['phc']) - np.int64(hst['dhc'])

    # previous frame (time tuple)
    prev_frm = (0,0)
    # previous frame index
    prev_i = -1
    # look up table to the auxilary frames from the regular ones
    timings = []
    fi2afi = []
    for i in range(len(t_frms)):
        # time point as an average between the frame end points
        tp = int(np.mean([t_frms[i][0],t_frms[i][1]]))
        # alternative (more accurate) average through centre of mass
        t0 = t_frms[i][0]
        t1 = t_frms[i][1]
        if t1>=hst['dur']: t1 = hst['dur']-1
        t = np.arange(t0,t1)
        tcm = np.sum(diff[t]*t)/np.sum(diff[t])
        # get the tuple of the equivalent count level frame
        frm = frame_position(hst, tcm, tr0=tr0, tr1=tr1, verbose=False)
        # form the LUT
        if frm!=prev_frm:
            prev_frm = frm
            prev_i += 1
            timings.append(list(frm))
        fi2afi.append(prev_i)
        if verbose:
            print 't[{}, {}]; tp={}, tcm={} => frm id:{}, timings:{}'.format(t_frms[i][0], t_frms[i][1], tp, tcm, fi2afi[-1], timings[-1])
    # form the list of auxilary dynamic frames of equivalent count level (as in Cref) for reconstruction
    mfrm = ['fluid'] + timings 
    return {'timings':mfrm, 'frame_idx':fi2afi}


def dynamic_timings(flist, offset=0):
    '''
    Get start and end frame timings from a list of dynamic PET frame definitions.
    Arguments:
    flist can be 1D list of time duration for each dynamic frame, e.g.: flist = [15, 15, 15, 15, 30, 30, 30, ...]
        or a 2D list of lists having 2 entries per definition: first for the number of repetitions and the other
        for the frame duration, e.g.: flist = ['def', [4, 15], [8, 30], ...], meaning 4x15s, then 8x30s, etc.
    offset adjusts for the start time (usually when prompts are strong enough over randoms)
    The output is a dictionary:
    out['timings'] = [[0, 15], [15, 30], [30, 45], [45, 60], [60, 90], [90, 120], [120, 150], ...]
    out['total'] = total time
    out['frames'] = array([ 15,  15,  15,  15,  30,  30,  30,  30, ...])

    '''
    if not isinstance(flist, list):
        raise TypeError('Wrong type of frame data input')
    if all([isinstance(t,(int, np.int32, np.int16, np.int8, np.uint8, np.uint16, np.uint32)) for t in flist]):
        tsum = offset
        # list of frame timings
        if offset>0:
            t_frames = ['timings', [0, offset]]
        else:
            t_frames = ['timings']
        for i in range(len(flist)):
            # frame start time
            t0 = tsum
            tsum += flist[i]
            # frame end time
            t1 = tsum
            # append the timings to the list
            t_frames.append([t0, t1])
        frms = np.uint16(flist)

    elif all( [isinstance(t,list) and len(t)==2 for t in flist[1:]] ) and flist[0]=='def':
        flist = flist[1:]
        if offset>0:
            flist.insert(0,[0,offset])
            farray = np.asarray(flist, dtype=np.uint16)
        else:
            farray = np.array(flist)
        # number of dynamic frames
        nfrm = np.sum(farray[:,0])
        # list of frame duration
        frms = np.zeros(nfrm,dtype=np.uint16)
        #frame iterator
        fi = 0
        #time sum of frames
        tsum = 0
        # list of frame timings
        t_frames = ['timings']
        for i in range(0, farray.shape[0]):
            for t in range(0, farray[i,0]):
                # frame start time
                t0 = tsum
                tsum += farray[i,1]
                # frame end time
                t1 = tsum
                # append the timings to the list
                t_frames.append([t0, t1])
                frms[fi] = farray[i,1]
                fi += 1
    else:
        raise TypeError('Unrecognised time frame definitions.')
    # prepare the output dictionary
    out = {'total':tsum, 'frames':frms, 'timings':t_frames}
    return out
#=================================================================================================


"""Image reconstruction from raw PET data"""
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"
#------------------------------------------------------------------------------

import numpy as np
import random
import sys
import os
import scipy.ndimage as ndi
from collections import namedtuple
from tqdm.auto import trange

import petprj

from niftypet.nipet.img import mmrimg
from niftypet.nipet import mmrnorm
from niftypet.nipet import mmraux
from niftypet import nimpa

# for isotope info
import resources

#reconstruction mode:
# 0 - no attenuation  and  no scatter
# 1 - attenuation  and   no scatter
# 2 - attenuation and scatter given as input parameter
# 3 - attenuation  and  scatter
recModeStr = ['_noatt_nosct_', '_nosct_', '_noatt_', '_', '_ute_']


# fwhm in [mm]
def fwhm2sig(fwhm, Cnt):
    return (0.1*fwhm/Cnt['SZ_VOXY']) / (2*(2*np.log(2))**.5)


#=========================================================================
# OSEM RECON
#-------------------------------------------------------------------------
def get_subsets14(n, params):
    ''' Define the n-th subset out of 14 in the transaxial projection space
    '''
    Cnt = params['Cnt']
    txLUT = params['txLUT']

    # just for check of sums (have to be equal for all subsets to make them balanced)
    aisum = np.sum(txLUT['msino'], axis=0)
    # number of subsets
    N = 14
    # projections per subset
    P = Cnt['NSANGLES']/N
    # the remaining projections which have to be spread over the N subsets with a given frequency
    fs = N/float(P-N)
    # generate sampling pattern for subsets up to N out of P
    sp = np.array([np.arange(i,Cnt['NSANGLES'],P) for i in range(N)])
    # ======================================
    S = np.zeros((N,P),dtype=np.int16)
    # ======================================
    # sum of sino angle projections
    totsum = np.zeros(N, dtype=np.int32)
    # iterate subset (which is also the angle iterator within block b)
    for s in range(N):
        # list of sino angular indexes for a given subset
        si = []
        #::::: iterate sino blocks.  This bit may be unnecessary, it can be taken directly from sp array
        for b in range(N):
            #--angle index within a sino block depending on subset s
            ai = (s+b)%N
            #--angle index for whole sino
            sai = sp[ai, b]
            si.append(sai)
            totsum[s] += aisum[sai]
        #:::::
        # deal with the remaining part, ie, P-N per block
        rai = np.int16( np.floor( np.arange(s,2*N,fs)[:4]%N ) )
        for i in range(P-N):
            sai = sp[-1,rai[i]]+i+1
            totsum[s] += aisum[sai]
            si.append(sai)
        # print si
        S[s] = np.array((si))

    # get the projection bin index for transaxial gpu sinos
    tmsk = txLUT['msino']>0
    Smsk = -1*np.ones(tmsk.shape, dtype=np.int32)
    Smsk[tmsk] = range(Cnt['Naw'])

    iprj = Smsk[:,S[n]]
    iprj = iprj[iprj>=0]

    # n=0; plot(S[n,:-4],ones(14), '*'); plot(S[n,-4:],ones(4), 'o')

    # Smsk = -1*np.ones(tmsk.shape, dtype=np.int32)
    # q=-1*ones(Cnt['Naw'])
    # q[iprj] = 3
    # Smsk[tmsk] = q

    return iprj, S
#---------------------------------------------------------------------------------------

#---------------------------------------------------------------------------------------------------------#
#=== OSEM image reconstruction with several modes (with/without scatter and/or attenuation correction) ===#
def osemone(
        datain,
        mumaps,
        hst,
        scanner_params,
        recmod=3,
        itr=4, 
        fwhm=0.,
        mask_radius=29.,
        sctsino=np.array([]),
        outpath='',
        store_img=False, frmno='', fcomment='',
        store_itr=[],
        emmskS=False,
        ret_sinos=False,
        attnsino = None,
        randsino = None,
        normcomp = None):


    #---------- sort out OUTPUT ------------
    #-output file name for the reconstructed image, initially assume n/a
    fout = 'n/a'
    if store_img or store_itr:
        if outpath=='':
            opth = os.path.join( datain['corepath'], 'reconstructed' )
        else:
            opth = outpath
        mmraux.create_dir(opth)
    #----------

    # Get particular scanner parameters: Constants, transaxial and axial LUTs
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    import time
    from niftypet import nipet
    # from niftypet.nipet.sct import mmrsct
    # from niftypet.nipet.prj import mmrhist

    if Cnt['VERBOSE']: print 'i> reconstruction in mode', recmod

    # get object and hardware mu-maps
    muh, muo = mumaps

    # get the GPU version of the image dims
    mus = mmrimg.convert2dev(muo+muh, Cnt)

    if Cnt['SPN']==1:
        snno = Cnt['NSN1']
    elif Cnt['SPN']==11:
        snno = Cnt['NSN11']

    # remove gaps from the prompt sino
    psng = mmraux.remgaps(hst['psino'], txLUT, Cnt)

    #=========================================================================
    # GET NORM
    #-------------------------------------------------------------------------
    if normcomp == None:
        ncmp, _ = mmrnorm.get_components(datain, Cnt)
    else:
        ncmp = normcomp
        print 'w> using user-defined normalisation components'
    nsng = mmrnorm.get_sinog(datain, hst, axLUT, txLUT, Cnt, normcomp=ncmp)
    #=========================================================================

    #=========================================================================
    # ATTENUATION FACTORS FOR COMBINED OBJECT AND BED MU-MAP
    #-------------------------------------------------------------------------
    #> combine attenuation and norm together depending on reconstruction mode
    if recmod==0:
        asng = np.ones(psng.shape, dtype=np.float32)
    else:
        #> check if the attenuation sino is given as an array
        if isinstance(attnsino, np.ndarray) \
                and attnsino.shape==(Cnt['NSN11'], Cnt['NSANGLES'], Cnt['NSBINS']):
            asng = mmraux.remgaps(attnsino, txLUT, Cnt)
            print 'i> using provided attenuation factor sinogram'
        elif isinstance(attnsino, np.ndarray) \
                and attnsino.shape==(Cnt['Naw'], Cnt['NSN11']):
            asng = attnsino
            print 'i> using provided attenuation factor sinogram'
        else:
            asng = np.zeros(psng.shape, dtype=np.float32)
            petprj.fprj(asng, mus, txLUT, axLUT, np.array([-1], dtype=np.int32), Cnt, 1)
    #> combine attenuation and normalisation
    ansng = asng*nsng
    #=========================================================================

    #=========================================================================
    # Randoms
    #-------------------------------------------------------------------------
    if isinstance(randsino, np.ndarray):
        rsino = randsino
        rsng = mmraux.remgaps(randsino, txLUT, Cnt)
    else:
        rsino, snglmap = nipet.lm.mmrhist.rand(hst['fansums'], txLUT, axLUT, Cnt)
        rsng = mmraux.remgaps(rsino, txLUT, Cnt)
    #=========================================================================

    #=========================================================================
    # SCAT
    #-------------------------------------------------------------------------
    if recmod==2:
        if sctsino.size>0:
            ssng = mmraux.remgaps(sctsino, txLUT, Cnt)
        elif sctsino.size==0 and os.path.isfile(datain['em_crr']):
            emd = nimpa.getnii(datain['em_crr'])
            ssn, sssr, amsk = nipet.sct.mmrsct.vsm(mumaps, emd['im'], datain, hst, rsino, 0.1, txLUT, axLUT, Cnt)
            ssng = mmraux.remgaps(ssn, txLUT, Cnt)
        else:
            print 'e> no emission image available for scatter estimation!  check if it''s present or the path is correct.'
            sys.exit()
    else:
        ssng = np.zeros(rsng.shape, dtype=rsng.dtype)
    #=========================================================================

    if Cnt['VERBOSE']:
        print '\n>------ OSEM (', itr,  ') -------\n'
    #------------------------------------
    Sn = 14 # number of subsets
    #-get one subset to get number of projection bins in a subset
    Sprj, s = get_subsets14(0,scanner_params)
    Nprj = len(Sprj)
    #-init subset array and sensitivity image for a given subset
    sinoTIdx = np.zeros((Sn, Nprj+1), dtype=np.int32)
    #-init sensitivity images for each subset
    imgsens = np.zeros((Sn, Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)
    for n in range(Sn):
        sinoTIdx[n,0] = Nprj #first number of projection for the given subset
        sinoTIdx[n,1:], s = get_subsets14(n,scanner_params)
        # sensitivity image
        petprj.bprj(imgsens[n,:,:,:], ansng[sinoTIdx[n,1:],:], txLUT, axLUT,  sinoTIdx[n,1:], Cnt )
    #-------------------------------------

    #-mask for reconstructed image.  anything outside it is set to zero
    msk = mmrimg.get_cylinder(Cnt, rad=mask_radius, xo=0, yo=0, unival=1, gpu_dim=True)>0.9

    #-init image
    img = np.ones((Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)

    #-decay correction
    lmbd = np.log(2)/resources.riLUT[Cnt['ISOTOPE']]['thalf']
    if Cnt['DCYCRR'] and 't0' in hst and 'dur' in hst:
        dcycrr = np.exp(lmbd*hst['t0'])*lmbd*hst['dur'] / (1-np.exp(-lmbd*hst['dur']))
        # apply quantitative correction to the image
        qf = ncmp['qf'] / resources.riLUT[Cnt['ISOTOPE']]['BF'] / float(hst['dur'])
        qf_loc = ncmp['qf_loc']
    elif not Cnt['DCYCRR'] and 't0' in hst and 'dur' in hst:
        dcycrr = 1.
        # apply quantitative correction to the image
        qf = ncmp['qf'] / resources.riLUT[Cnt['ISOTOPE']]['BF'] / float(hst['dur'])
        qf_loc = ncmp['qf_loc']
    else:
        dcycrr = 1.
        qf = 1.
        qf_loc = 1.

    #-affine matrix for the reconstructed images
    B = mmrimg.image_affine(datain, Cnt)

    #-time it
    stime = time.time()

    # import pdb; pdb.set_trace()

    #=========================================================================
    # OSEM RECONSTRUCTION
    #-------------------------------------------------------------------------
    for k in trange(itr, disable=not Cnt['VERBOSE'], desc="OSEM"):
        petprj.osem(img, msk, psng, rsng, ssng, nsng, asng, imgsens, txLUT, axLUT, sinoTIdx, Cnt)
        if np.nansum(img)<0.1:
            print '---------------------------------------------------------------------'
            print 'w> it seems there is not enough true data to render reasonable image.'
            print '---------------------------------------------------------------------'
            #img[:]=0
            itr = k
            break
        if recmod>=3 and ( ((k<itr-1) and (itr>1)) ): # or (itr==1)
            sct_time = time.time()
            ssn, sssr, amsk = nipet.sct.mmrsct.vsm(
                mumaps,
                mmrimg.convert2e7(img, Cnt),
                datain,
                hst,
                rsino,
                scanner_params,
                prcntScl=0.1,
                emmsk=emmskS)
            ssng = mmraux.remgaps(ssn, txLUT, Cnt)
            if Cnt['VERBOSE']: print 'i> scatter time:', (time.time() - sct_time)
        # save images during reconstruction if requested
        if store_itr and k in store_itr:
            im = mmrimg.convert2e7(img * (dcycrr*qf*qf_loc), Cnt)
            fout =  os.path.join(opth, os.path.basename(datain['lm_bf'])[:8] \
                + frmno +'_t'+str(hst['t0'])+'-'+str(hst['t1'])+'sec' \
                +'_itr'+str(k)+fcomment+'_inrecon.nii.gz')
            nimpa.array2nii( im[::-1,::-1,:], B, fout)


    if Cnt['VERBOSE']: print 'i> recon time:', (time.time() - stime)
    #=========================================================================


    if Cnt['VERBOSE']:
        print 'i> applying decay correction of', dcycrr
        print 'i> applying quantification factor', qf, 'to the whole image for the frame duration of :', hst['dur']

    img *= dcycrr * qf * qf_loc #additional factor for making it quantitative in absolute terms (derived from measurements)

    #---- save images -----
    #-first convert to standard mMR image size
    im = mmrimg.convert2e7(img, Cnt)

    #-description text to NIfTI
    #-attenuation number: if only bed present then it is 0.5
    attnum =  ( 1*(np.sum(muh)>0.5)+1*(np.sum(muo)>0.5) ) / 2.
    descrip =   'alg=osem'+ \
                ';sub=14'+ \
                ';att='+str(attnum*(recmod>0))+ \
                ';sct='+str(1*(recmod>1))+ \
                ';spn='+str(Cnt['SPN'])+ \
                ';itr='+str(itr) +\
                ';fwhm='+str(fwhm) +\
                ';t0='+str(hst['t0']) +\
                ';t1='+str(hst['t1']) +\
                ';dur='+str(hst['dur']) +\
                ';qf='+str(qf)

    if fwhm>0:
        im = ndi.filters.gaussian_filter(im, fwhm2sig(fwhm, Cnt), mode='mirror')
    if store_img:
        fout =  os.path.join(opth, os.path.basename(datain['lm_bf'])[:8] \
                + frmno +'_t'+str(hst['t0'])+'-'+str(hst['t1'])+'sec' \
                +'_itr'+str(itr)+fcomment+'.nii.gz')
        if Cnt['VERBOSE']: print 'i> saving image to: ', fout
        nimpa.array2nii( im[::-1,::-1,:], B, fout, descrip=descrip)


    # returning:
    # (0) E7 image [can be smoothed];
    # (1) file name of saved E7 image
    # (2) [optional] scatter sino
    # (3) [optional] single slice rebinned scatter
    # (4) [optional] mask for scatter scaling based on attenuation data
    # (5) [optional] random sino
    # if ret_sinos and recmod>=3:
    #     recout = namedtuple('recout', 'im, fpet, ssn, sssr, amsk, rsn')
    #     recout.im   = im
    #     recout.fpet = fout
    #     recout.ssn  = ssn
    #     recout.sssr = sssr
    #     recout.amsk = amsk
    #     recout.rsn  = rsino
    # else:
    #     recout = namedtuple('recout', 'im, fpet')
    #     recout.im   = im
    #     recout.fpet = fout
    if ret_sinos and recmod>=3 and itr>1:
        RecOut = namedtuple('RecOut', 'im, fpet, affine, ssn, sssr, amsk, rsn')
        recout = RecOut(im, fout, B, ssn, sssr, amsk, rsino)
    else:
        RecOut = namedtuple('RecOut', 'im, fpet, affine')
        recout = RecOut(im, fout, B)

    return recout










#===============================================================================
# EMML
# def emml(   datain, mumaps, hst, txLUT, axLUT, Cnt,
#             recmod=3, itr=10, fwhm=0., mask_radius=29., store_img=True, ret_sinos=False, sctsino = None, randsino = None, normcomp = None):

#     #subsets (when not used)
#     sbs = np.array([-1], dtype=np.int32)

#     # get object and hardware mu-maps
#     muh, muo = mumaps

#     # get the GPU version of the image dims
#     mus = mmrimg.convert2dev(muo+muh, Cnt)

#     # remove gaps from the prompt sinogram
#     psng = mmraux.remgaps(hst['psino'], txLUT, Cnt)

#     #=========================================================================
#     # GET NORM
#     #-------------------------------------------------------------------------
#     if normcomp == None:
#         ncmp, _ = mmrnorm.get_components(datain, Cnt)
#     else:
#         ncmp = normcomp
#         print 'w> using user-defined normalisation components'
#     nrmsng = mmrnorm.get_sinog(datain, hst, axLUT, txLUT, Cnt, normcomp=ncmp)
#     #=========================================================================


#     #=========================================================================
#     # Randoms
#     #-------------------------------------------------------------------------
#     if randsino == None:
#         rsino, snglmap = mmrhist.rand(hst['fansums'], txLUT, axLUT, Cnt)
#         rsng = mmraux.remgaps(rsino, txLUT, Cnt)
#     else:
#         rsino = randsino
#         rsng = mmraux.remgaps(randsino, txLUT, Cnt)
#     #=========================================================================


#     #=========================================================================
#     # ATTENUATION FACTORS FOR COMBINED OBJECT AND BED MU-MAP
#     #-------------------------------------------------------------------------
#     # combine attenuation and norm together depending on reconstruction mode
#     if recmod==0:
#         asng = np.ones(psng.shape, dtype=np.float32)
#     else:
#         asng = np.zeros(psng.shape, dtype=np.float32)
#         petprj.fprj(asng, mus, txLUT, axLUT, sbs, Cnt, 1)
#     attnrmsng = asng*nrmsng
#     #=========================================================================


#     #=========================================================================
#     # SCATTER and the additive term
#     #-------------------------------------------------------------------------
#     if recmod==2:
#         if sctsino != None:
#             # remove the gaps from the provided scatter sinogram
#             ssng = mmraux.remgaps(sctsino, txLUT, Cnt)
#         elif sctsino == None and os.path.isfile(datain['em_crr']):
#             # estimate scatter from already reconstructed and corrected emission image
#             emd = nimpa.prc.getnii(datain['em_crr'], Cnt)
#             ssn, sssr, amsk = mmrsct.vsm(mumaps, emd['im'], datain, hst, rsn, 0.1, txLUT, axLUT, Cnt)
#             ssng = mmraux.remgaps(ssn, txLUT, Cnt)
#         else:
#             print 'e> no emission image available for scatter estimation!  check if it''s present or the path is correct.'
#             sys.exit()
#     else:
#         ssng = np.zeros(rsng.shape, dtype=rsng.dtype)
#     # form the additive term
#     rssng = (rsng + ssng) / attnrmsng
#     #=========================================================================

#     #mask for reconstructed image
#     msk = mmrimg.get_cylinder(Cnt, rad=mask_radius, xo=0, yo=0, unival=1, gpu_dim=True)>0.9
#     # estimated image
#     imrec = np.ones((Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)
#     # backprj image
#     bim = np.zeros((Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)
#     # Get sensitivity image by backprojection
#     sim = np.zeros((Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)
#     petprj.bprj(sim, attnrmsng, txLUT, axLUT, sbs, Cnt)
#     #init estimate sino
#     esng = np.zeros((Cnt['Naw'], Cnt['NSN11']), dtype=np.float32)


#     for k in range(itr):
#         print '>--------- ITERATION', k, '-----------<'
#         esng[:] = 0
#         petprj.fprj(esng, imrec, txLUT, axLUT, sbs, Cnt, 0)
#         # esng *= attnrmsng
#         esng += (rssng+ssng)
#         # crr = attnrmsng*(psng/esng)
#         crr = psng/esng
#         bim[:] = 0
#         petprj.bprj(bim, crr, txLUT, axLUT,  sbs, Cnt)
#         bim /= sim
#         imrec *= msk*bim
#         imrec[np.isnan(imrec)] = 0

#         if recmod>=3 and ( ((k<itr-1)and(itr>1))):
#             sct_time = time.time()
#             ssn, sssr, amsk = mmrsct.vsm(mumaps, mmrimg.convert2e7(img, Cnt), datain, hst, rsn, scanner_params, prcntScl=0.1, emmsk=emmskS)
#             ssng = mmraux.remgaps(ssn, txLUT, Cnt) / attnrmsng
#             if Cnt['VERBOSE']: print 'i> scatter time:', (time.time() - sct_time)

#     # decay correction
#     lmbd = np.log(2)/resources.riLUT[Cnt['ISOTOPE']]['thalf']
#     dcycrr = np.exp(lmbd*hst['t0'])*lmbd*hst['dur'] / (1-np.exp(-lmbd*hst['dur']))
#     # apply quantitative correction to the image
#     qf = ncmp['qf'] / resources.riLUT[Cnt['ISOTOPE']]['BF'] / float(hst['dur'])
#     if Cnt['VERBOSE']: print 'i> applying quantification factor', qf, 'to the whole image for the frame duration of :', hst['dur']
#     imrec *= dcycrr * qf * 0.205 #additional factor for making it quantitative in absolute terms (derived from measurements)

#     # convert to standard mMR image size
#     im = mmrimg.convert2e7(imrec, Cnt)

#     if fwhm>0:
#         im = ndi.filters.gaussian_filter(im, fwhm2sig(fwhm, Cnt), mode='mirror')

#     #save images
#     B = mmrimg.image_affine(datain, Cnt)
#     fout = ''

#     if store_img:
#         # description text to NIfTI
#         # attenuation number: if only bed present then it is 0.5
#         attnum =  ( 1*(np.sum(muh)>0.5)+1*(np.sum(muo)>0.5) ) / 2.
#         descrip =   'alg=emml'+ \
#                     ';sub=0'+ \
#                     ';att='+str(attnum*(recmod>0))+ \
#                     ';sct='+str(1*(recmod>1))+ \
#                     ';spn='+str(Cnt['SPN'])+ \
#                     ';itr='+str(itr)+ \
#                     ';fwhm='+str(fwhm) +\
#                     ';t0='+str(hst['t0']) +\
#                     ';t1='+str(hst['t1']) +\
#                     ';dur='+str(hst['dur']) +\
#                     ';qf='+str(qf)
#         fout =  os.path.join(datain['corepath'], os.path.basename(datain['lm_dcm'])[:8]+'_emml_'+str(itr)+'.nii.gz')
#         nimpa.array2nii( im[::-1,::-1,:], B, fout, descrip=descrip)

#     if ret_sinos and recmod>=3 and itr>1:
#         RecOut = namedtuple('RecOut', 'im, fpet, affine, ssn, sssr, amsk, rsn')
#         recout = RecOut(im, fout, B, ssn, sssr, amsk, rsn)
#     else:
#         RecOut = namedtuple('RecOut', 'im, fpet, affine')
#         recout = RecOut(im, fout, B)

#     return recout





#=============================================================================
# OSEM

# def osem14(datain, mumaps, hst, txLUT, axLUT, Cnt,
#             recmod=3, itr=4, fwhm=0., mask_radius=29.):

#     muh, muo = mumaps
#     mus = mmrimg.convert2dev(muo+muh, Cnt)

#     if Cnt['SPN']==1:
#         snno = Cnt['NSN1']
#     elif Cnt['SPN']==11:
#         snno = Cnt['NSN11']

#     #subsets (when not used)
#     sbs = np.array([-1], dtype=np.int32)

#     # remove gaps from the prompt sino
#     psng = mmraux.remgaps(hst['psino'], txLUT, Cnt)

#     #=========================================================================
#     # GET NORM
#     #-------------------------------------------------------------------------
#     nrmsng = mmrnorm.get_sinog(datain, hst, axLUT, txLUT, Cnt)
#     #=========================================================================

#     #=========================================================================
#     # RANDOMS ESTIMATION
#     #-------------------------------------------------------------------------
#     rsino, snglmap = mmrhist.rand(hst['fansums'], txLUT, axLUT, Cnt)
#     rndsng = mmraux.remgaps(rsino, txLUT, Cnt)
#     #=========================================================================

#     #=========================================================================
#     # FORM THE ADDITIVE TERM
#     #-------------------------------------------------------------------------
#     if recmod==0 or recmod==1 or recmod==3 or recmod==4:
#         rssng = rndsng
#     elif recmod==2:
#         if os.path.isfile(datain['em_crr']):
#             emd = nimpa.getnii(datain['em_crr'])
#             ssn, sssr, amsk = mmrsct.vsm(mumaps, emd['im'], datain, hst, rsino, 0.1, txLUT, axLUT, Cnt)
#             rssng = rndsng + mmraux.remgaps(ssn, txLUT, Cnt)
#         else:
#             print 'e> no emission image availble for scatter estimation!  check if it''s present or the path is correct.'
#             sys.exit()
#     #=========================================================================

#     #=========================================================================
#     # ATTENUATION FACTORS FOR COMBINED OBJECT AND BED MU-MAP
#     #-------------------------------------------------------------------------
#     # combine attenuation and norm together depending on reconstruction mode
#     if recmod==0 or recmod==2:
#         attnrmsng = nrmsng
#     else:
#         attnrmsng = np.zeros(psng.shape, dtype=np.float32)
#         petprj.fprj(attnrmsng, mus, txLUT, axLUT, sbs, Cnt, 1)
#         attnrmsng *= nrmsng
#     #=========================================================================

#     #mask for reconstructed image
#     rcnmsk = mmrimg.get_cylinder(Cnt, rad=mask_radius, xo=0, yo=0, unival=1, gpu_dim=True)
#     #-------------------------------------------------------------------------
#     # number of subsets
#     Sn = 14
#     # get one subset to get number of projection bins in a subset
#     Sprj, s = get_subsets14(0,txLUT,Cnt)
#     # init subset array and sensitivity image for a given subset
#     sinoTIdx = np.zeros((Sn, len(Sprj)), dtype=np.int32)
#     sim = np.zeros((Sn, Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)
#     for n in range(Sn):
#         sinoTIdx[n,:], s = get_subsets14(n,txLUT,Cnt)
#         petprj.bprj(sim[n,:,:,:], attnrmsng, txLUT, axLUT, sinoTIdx[n,:], Cnt)
#     #--------------------------------------------------------------------------

#     # estimated image
#     xim = np.ones((Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)
#     # backprj image
#     bim = np.ones((Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)
#     # init scatter sino (zeros)
#     ssng  = np.zeros((Cnt['Naw'],   snno), dtype=np.float32)
#     # sinogram subset mask
#     sbmsk = np.zeros((txLUT['Naw'], snno), dtype=np.bool)
#     # estimated sinogram (forward model)
#     esng  = np.zeros((txLUT['Naw'], snno), dtype=np.float32)

#     for k in range(itr):
#         # randomly go through subsets + ssng
#         sn = range(Sn)
#         random.shuffle(sn)
#         s = 0
#         for n in sn:
#             print ' '
#             print k, '>--------- SUBSET', s, n, '-----------'
#             s+=1
#             sbmsk[:] = False
#             sbmsk[sinoTIdx[n,:],:] = True
#             esng[:] = 0
#             petprj.fprj(esng, xim, txLUT, axLUT, sinoTIdx[n,:], Cnt, 0)
#             esng *= attnrmsng
#             if (recmod==3  or recmod==4):
#                 esng += (rssng+ssng)*sbmsk
#             else:
#                 esng += rssng*sbmsk

#             # corrections to be backprojected to the image space
#             crr = attnrmsng*(np.float32(psng)/esng)
#             crr[np.isnan(crr)] = 0
#             crr[np.isinf(crr)] = 0
#             petprj.bprj(bim, crr, txLUT, axLUT, sinoTIdx[n,:], Cnt)
#             # devide the backprojected image by the corresponding subset sensitivity image
#             bim /= sim[n,:,:,:]
#             # apply the reconstruction mask
#             xim *= rcnmsk*bim
#             # get rid of any NaN values, if any
#             xim[np.isnan(xim)]=0

#             # plt.figure(); plt.imshow(xim[:,:,70], interpolation='none', cmap='gray'); plt.show()

#         #plt.figure(); plt.imshow(xim[:,:,70], interpolation='none', cmap='gray'); plt.show()
#         if (recmod==3  or recmod==4) and k<itr-1:
#             ssn, sssr, amsk = mmrsct.vsm(mumaps, mmrimg.convert2e7(xim, Cnt), datain, hst, rsino, txLUT, axLUT, Cnt, prcntScl=0.1, emmsk=True)
#             ssng = mmraux.remgaps(ssn, txLUT, Cnt)

#     #---- save images -----
#     #first convert to standard mMR image size
#     im = mmrimg.convert2e7(xim, Cnt)
#     B = mmrimg.image_affine(datain, Cnt)
#     #save the nii image
#     fout = os.path.dirname(datain['lm_dcm'])+'/'+os.path.basename(datain['lm_dcm'])[:8]+'_osem14_i'+str(itr)+'_s'+str(Cnt['SPN'])+'_r'+str(recmod)+'.nii'
#     nimpa.array2nii( im[::-1,::-1,:], B, fout)
#     #do smoothing and save the image
#     if fwhm>0:
#         imsmo = ndi.filters.gaussian_filter(im, fwhm2sig(fwhm, Cnt), mode='mirror')
#         nimpa.array2nii( imsmo[::-1,::-1,:], B,
#             os.path.dirname(datain['lm_dcm'])+'/'+os.path.basename(datain['lm_dcm'])[:8]+'_osem14_i'+str(itr)+'_s'+str(Cnt['SPN'])+'_r'+str(recmod)+'_smo'+str(fwhm)+'.nii')

#     if recmod==3:
#         datain['em_crr'] = fout

#     return im, fout
